/* creating object types for path:
   
   PATH (T_PATH): object which consist of multiple points (T_POINT)    
   +--------------+
   |     T_PATH   |           +----------+
   +--------------+ <>--------| T_POINT  |
   | get_length() |           +----------+
   +--------------+           |   x      |
                              |   y      |
                              +----------+
    
    get_length:   calculates and sums distanses between points
*/

create type t_point as object (
  x float,
  y float
);
/

create type t_point_list is table of t_point;
/

create type t_path as object (
  point_list t_point_list,
  member function get_length return float
);
/

create or replace
TYPE BODY T_PATH AS

  member function get_length return float AS
    v_len float := 0;
  BEGIN
    /* TODO implementation required */
    --RETURN NULL;
    for i in 1..self.point_list.count-1
    loop
      v_len := v_len + sqrt( (self.point_list(i).x-self.point_list(i+1).x)*(self.point_list(i).x-self.point_list(i+1).x) +
                             (self.point_list(i).y-self.point_list(i+1).y)*(self.point_list(i).y-self.point_list(i+1).y) );
    end loop;
    return v_len;
  END get_length;

END;
/


--=========================================================================


/*
 create table, which have path column of T_PATH type 
*/

create table paths (
  path_name varchar2(200),
  path_desc varchar2(1000),
  path t_path 
) nested table path.point_list store as t_path_nt;
/


--- insert sample data pt paths table
insert into paths values (
  'path1', 
  'test path',
  t_path(t_point_list(t_point(0,0), t_point(0, 23), t_point(10, 44)))
);

commit;

select t_path.get_length(self => path), p.* from paths p;

--=========================================================================
/* 
  create operator and operator implementation function
  operator PATH_BETWEEN(x, a, b) return 1 or 0
    id path length (path.get_length()) is between a and b return 1 else 0.
*/

create or replace
function path_between_fn (path t_path, f1 float, f2 float) return integer is 
begin
  if path.get_length between f1 and f2 then
    return 1;
  else 
    return 0;
  end if;
end;
/

create or replace operator path_between binding (t_path, float, float) return integer 
using path_between_fn;
/

--=========================================================================

--create type to store list of ROWIDs
create or replace type t_rowid_list is table of varchar2(200);
/

--create sequence to generate new tables for index storages
create sequence sqn_idxtable_id nocycle;


--create index "metadata" table to link user defined index name with index storage
create table idx_tables (
  idx_schema      varchar2(35),
  idx_name        varchar2(35),
  idx_table_name  varchar2(200)
);

--=========================================================================
/* create indextype implementation object type */

-- object type definition
create or replace
type PathIndexTypeMethods AUTHID CURRENT_USER as object  (
  --- sql fetch cursor number
  curnum integer, 
  
  static function ODCIGetInterfaces(ifclist OUT SYS.ODCIObjectList) RETURN NUMBER,
  
  -- index definition functions
  static function ODCIIndexCreate(ia SYS.ODCIIndexInfo, parm varchar2, 
    env SYS.ODCIEnv) return number,
  STATIC FUNCTION ODCIIndexAlter(ia SYS.ODCIIndexInfo, parms IN OUT VARCHAR2, 
    alter_option NUMBER, env sys.ODCIEnv) return number,
  STATIC FUNCTION ODCIIndexDrop(ia SYS.ODCIIndexInfo, env sys.ODCIEnv) return number,
  
  -- index maintenance functions
  static FUNCTION ODCIIndexInsert(ia sys.ODCIIndexInfo, rid VARCHAR2, 
    newval t_path, env sys.ODCIEnv) return number,
  static FUNCTION ODCIIndexInsert(ia sys.ODCIIndexInfo, ridlist sys.ODCIRidList, 
    newvallist t_float_list, env sys.ODCIEnv) return number,
  static FUNCTION ODCIIndexDelete(ia sys.ODCIIndexInfo, rid VARCHAR2, 
    oldval t_path, env sys.ODCIEnv) RETURN NUMBER,
  static FUNCTION ODCIIndexUpdate(ia sys.ODCIIndexInfo, rid VARCHAR2, 
    oldval t_path, newval t_path, env sys.ODCIEnv) RETURN NUMBER,

  -- index scan methods
  static FUNCTION ODCIIndexStart( sctx IN OUT PathIndexTypeMethods, 
    ia sys.ODCIIndexInfo, pi sys.ODCIPredInfo, qi sys.ODCIQueryInfo, strt integer, 
    stop integer, f1 float, f2 float, env sys.ODCIEnv) RETURN NUMBER,
  member FUNCTION ODCIIndexFetch(nrows IN NUMBER, rids OUT sys.ODCIRidList, 
    env sys.ODCIEnv) RETURN NUMBER,
  member FUNCTION ODCIIndexClose(env sys.ODCIEnv) RETURN NUMBER
);
/

-- object type body
create or replace
TYPE BODY PATHINDEXTYPEMETHODS 
 AS 
  
  static function ODCIGetInterfaces(ifclist OUT SYS.ODCIObjectList) RETURN NUMBER AS
  BEGIN
    ifclist := sys.ODCIObjectList(sys.ODCIObject('SYS','ODCIINDEX2'));
    RETURN ODCIConst.Success;
  END ODCIGetInterfaces;

  static function ODCIIndexCreate(
    ia SYS.ODCIIndexInfo, 
    parm varchar2, 
    env SYS.ODCIEnv) 
  return number   AS
    v_cnt integer;
    v_cnt2 integer;
    v_idx_sqn integer;   
    idx_tab_name varchar2(50);
    nested_tab_name varchar2(50);
    v_cur sys_refcursor;
    v_path T_PATH;
    v_rowid varchar2(200);
    v_rowid2 varchar2(200);
  BEGIN
    --- 1. check num of indexed columns. only one column can be indexed
    if ia.IndexCols.count > 1 then
      return ODCIConst.Error;
    end if;
    
    --- 1. check indexed column type. only T_PATH type is accepted
    if upper(ia.IndexCols(1).ColTypeName) <> 'T_PATH'  then
      return ODCIConst.Error;
    end if;
    
    v_idx_sqn := sqn_idxtable_id.nextval;
    idx_tab_name := 'IDX$'||v_idx_sqn||'TAB';
    nested_tab_name := 'RWS$'||v_idx_sqn||'TAB';
    
    --- 2. insert metadata about index
    insert into idx_tables values (ia.IndexSchema, ia.IndexName, idx_tab_name);
    
    --- 3. create IOT table to hold index data
    execute immediate 
      'CREATE TABLE '||ia.IndexSchema||'.'||idx_tab_name||' (
          path_len float primary key, 
          rowids t_rowid_list)
      organization index
      nested table rowids store as '||nested_tab_name;
  
    --- 4. check if table is empty, if not empty create index data
    execute immediate 
      'select count(0) from '||ia.IndexCols(1).TableSchema||'.'||ia.IndexCols(1).TableName||' where rownum < 2' into  v_cnt;
    if v_cnt > 0 then
      open v_cur for 'select rowid, '||ia.IndexCols(1).ColName||' from '||ia.IndexCols(1).TableSchema||'.'||ia.IndexCols(1).TableName;
      loop
        fetch v_cur into v_rowid, v_path;
        exit when v_cur%notfound;
        execute immediate 
          'select count(0), max(rowid) from '||idx_tab_name||' where path_len between round(:a, 6) -0.000001 and round(:a, 6) +0.000001' 
            into v_cnt2, v_rowid2 using v_path.get_length, v_path.get_length;
        if v_cnt2 = 1 then
          execute immediate 
            'insert into table(select rowids from '||idx_tab_name||' i where i.rowid = :a ) values (:b)'
              using v_rowid2, v_rowid; 
        else
          execute immediate 
            'insert into '||idx_tab_name||' values (:a, :b)'
              using v_path.get_length, t_rowid_list(v_rowid); 
        end if;
        commit;
      end loop;
    end if;
    return ODCIConst.Success;
  exception
    when others then
      dbms_output.put_line(dbms_utility.format_error_backtrace);
      dbms_output.put_line(dbms_utility.format_error_stack);
      return ODCIConst.Error;
  END ODCIIndexCreate;

  STATIC FUNCTION ODCIIndexAlter(ia SYS.ODCIIndexInfo, parms IN OUT VARCHAR2, 
    alter_option NUMBER, env sys.ODCIEnv) return number AS
  BEGIN
    /* TODO implementation required */
    return null;
  END ODCIIndexAlter;

  STATIC FUNCTION ODCIIndexDrop(ia SYS.ODCIIndexInfo, env sys.ODCIEnv) return number AS
    v_idx_tabname varchar2(200);
  BEGIN
    /* TODO implementation required */
    select idx_table_name into v_idx_tabname from idx_tables 
      where idx_name = ia.IndexName and idx_schema  = ia.IndexSchema;
    begin
      execute immediate 
        'drop table '||ia.IndexSchema||'.'||v_idx_tabname||' purge';
    exception
      when others then 
        null;
    end;
    delete from idx_tables 
        where idx_name = ia.IndexName and idx_schema  = ia.IndexSchema;
    commit;
    return ODCIConst.Success;
  END ODCIIndexDrop;

  static FUNCTION ODCIIndexInsert(ia sys.ODCIIndexInfo, rid VARCHAR2, 
    newval t_path, env sys.ODCIEnv)
  return number AS
    v_idx_table varchar2(50);  
    v_cnt integer;
    v_rowid varchar2(200);
  BEGIN
    /* TODO implementation required */
    select idx_table_name into v_idx_table from idx_tables 
      where idx_schema=ia.IndexSchema and idx_name=ia.IndexName;
    execute immediate 
      'select count(0), max(rowid) from '||v_idx_table||' where path_len between round(:a, 6)-0.000001 and round(:a, 6)+0.000001' 
        into v_cnt, v_rowid using newval.get_length, newval.get_length;
    if v_cnt = 1 then
      execute immediate 
        'insert into table(select rowids from '||v_idx_table||' i where i.rowid = :a) values (:b)'
          using v_rowid, rid; 
    else
      execute immediate 
        'insert into '||v_idx_table||' values (:a, :b)'
          using newval.get_length, t_rowid_list(rid); 
    end if;
    RETURN ODCIConst.Success;
  exception 
    when others then
      dbms_output.put_line(dbms_utility.format_error_backtrace);
      dbms_output.put_line(dbms_utility.format_error_stack);
      return ODCIConst.Error;
  END ODCIIndexInsert;

  static FUNCTION ODCIIndexInsert(ia sys.ODCIIndexInfo, ridlist sys.ODCIRidList, 
    newvallist t_float_list, env sys.ODCIEnv) 
  return number AS
    v_idx_table varchar2(50);  
    v_cnt integer;
    v_rowid varchar2(200);
  BEGIN
    /* TODO implementation required */
    select idx_table_name into v_idx_table from idx_tables 
      where idx_schema=ia.IndexSchema and idx_name=ia.IndexName;
    
    for i in 1..newvallist.count 
    loop
      execute immediate 
        'select count(0), max(rowid) from '||v_idx_table||' where path_len between round(:a, 6)-0.000001 and round(:a, 6)+0.000001' 
          into v_cnt, v_rowid using newvallist(i).get_length, newvallist(i).get_length;
      if v_cnt = 1 then
        execute immediate 
          'insert into table(select rowids from '||v_idx_table||' i where i.rowid = :a) values (:b)'
            using v_rowid, ridlist(i); 
      else
        execute immediate 
          'insert into '||v_idx_table||' values (:a, :b)'
            using newvallist(i).get_length, t_rowid_list(ridlist(i)); 
      end if;
    end loop;
  exception 
    when others then
      dbms_output.put_line(dbms_utility.format_error_backtrace);
      dbms_output.put_line(dbms_utility.format_error_stack);
      return ODCIConst.Error;  
  END ODCIIndexInsert;

  static FUNCTION ODCIIndexDelete(ia sys.ODCIIndexInfo, rid VARCHAR2, 
    oldval t_path, env sys.ODCIEnv) 
  RETURN NUMBER AS
    v_idx_table varchar2(50);  
    v_cnt integer;
    v_rowid varchar2(200);
  BEGIN
    /* TODO implementation required */
    select idx_table_name into v_idx_table from idx_tables 
      where idx_schema=ia.IndexSchema and idx_name=ia.IndexName;
    
    execute immediate 
      'select count(0), max(rowid)  from '||v_idx_table||' where path_len between round(:a, 6)-0.000001 and round(:a, 6)+0.000001' 
        into v_cnt, v_rowid using oldval.get_length, oldval.get_length;
    
    if v_cnt = 1 then
      execute immediate 
        'delete from table(select rowids from '||v_idx_table||' i where i.rowid = :a) where column_value = :b'
          using v_rowid, rid; 
    else
      execute immediate 
        'delete from '||v_idx_table||' where rowid = :a'
          using v_rowid; 
    end if;
    RETURN ODCIConst.Success;
  exception 
    when others then
      dbms_output.put_line(dbms_utility.format_error_backtrace);
      dbms_output.put_line(dbms_utility.format_error_stack);
      return ODCIConst.Error;
  END ODCIIndexDelete;

  static FUNCTION ODCIIndexUpdate(ia sys.ODCIIndexInfo, rid VARCHAR2, 
    oldval t_path, newval t_path, env sys.ODCIEnv) 
  RETURN NUMBER AS
    v_idx_table varchar2(50);  
    v_cnt_d integer;
    v_cnt_i integer;
    v_rowid_d varchar2(200);
    v_rowid_i varchar2(200);
  BEGIN
    /* TODO implementation required */
     select idx_table_name into v_idx_table from idx_tables 
      where idx_schema=ia.IndexSchema and idx_name=ia.IndexName;
    
    execute immediate 
      'select count(0), max(rowid) from '||v_idx_table||' where path_len between round(:a, 6)-0.000001 and round(:a, 6)+0.000001' 
        into v_cnt_d, v_rowid_d using oldval.get_length, oldval.get_length;
      
    execute immediate 
      'select count(0), max(rowid) from '||v_idx_table||' where path_len between round(:a, 6)-0.000001 and round(:a, 6)+0.000001'
        into v_cnt_i, v_rowid_i using newval.get_length, newval.get_length;  
    
    if v_cnt_d = 1 then
      execute immediate 
        'delete from table(select rowids from '||v_idx_table||' i where i.rowid = :a) where column_value = :b'
          using v_rowid_d, rid; 
    else
      execute immediate 
        'delete from '||v_idx_table||' i where i.rowid = :a'
          using v_rowid_d; 
    end if;
    
    if v_cnt_i = 1 then
      execute immediate 
        'insert into table(select rowids from '||v_idx_table||' i where i.rowid = :a) values (:b)'
          using v_rowid_i, rid; 
    else
      execute immediate 
        'insert into '||v_idx_table||' values (:a, :b)'
          using newval.get_length, t_rowid_list(rid); 
    end if;
    
    RETURN ODCIConst.Success;
  exception 
    when others then
      dbms_output.put_line(dbms_utility.format_error_backtrace);
      dbms_output.put_line(dbms_utility.format_error_stack);
      return ODCIConst.Error;
  END ODCIIndexUpdate;

  static FUNCTION ODCIIndexStart( sctx IN OUT PathIndexTypeMethods, 
    ia sys.ODCIIndexInfo, pi sys.ODCIPredInfo, qi sys.ODCIQueryInfo, strt integer, 
    stop integer, f1 float, f2 float, env sys.ODCIEnv) 
  RETURN NUMBER AS
    v_idx_table varchar2(200);  
    v_cnum integer;
    v_rowids varchar2(200);
    v_nrow integer;
    v_stmt varchar2(1000);
  BEGIN
    /* TODO implementation required */
    
    dbms_output.put_line('ODCIIndexStart>>>>>');
    sys.ODCIIndexInfoDump(ia);
    sys.ODCIPredInfoDump(pi);
    dbms_output.put_line('start key : '||strt);
    dbms_output.put_line('stop key : '||stop);
    dbms_output.put_line('f1 : '||f1);
    dbms_output.put_line('f2 : '||f2);
    
    dbms_output.put_line(pi.flags);
    
    select idx_table_name into v_idx_table from idx_tables 
        where idx_schema=ia.IndexSchema and idx_name=ia.IndexName;
    
    if strt = 1 then
      v_stmt  := 'select t.column_value from '
                 ||ia.IndexSchema||'.'||v_idx_table||
                 ' i, table(i.rowids) t where path_len between '
                 ||f1||' and '||f2;
    elsif strt = 0 then 
      v_stmt  := 'select t.column_value from '
                 ||ia.IndexSchema||'.'||v_idx_table||
                 ' i, table(i.rowids) t where path_len not between '
                 ||f1||' and '||f2;
    else 
      dbms_output.put_line('operator ret value wrong!');
      --raise_application_error(-20101, 'Unsupported operator');
      RETURN ODCIConst.error;
    end if;
    
    dbms_output.put_line(v_stmt);
    
    --if pi.flags = ODCIConst.PredExactMatch then
      v_cnum := dbms_sql.open_cursor;
      dbms_sql.parse(v_cnum, v_stmt, dbms_sql.native);
      dbms_sql.define_column(v_cnum, 1, v_rowids,200);
      v_nrow := dbms_sql.execute(v_cnum);
      sctx := pathindextypemethods(v_cnum);
    --else
    --  dbms_output.put_line('operator wrong!');
    --  raise_application_error(-20101, 'Unsupported operator');
    --  RETURN ODCIConst.error;
    --end if;
    RETURN ODCIConst.success;
  exception
    when others then
      dbms_output.put_line(dbms_utility.format_error_backtrace);
      dbms_output.put_line(dbms_utility.format_error_stack);
      return ODCIConst.Error;
  END ODCIIndexStart;

  member FUNCTION ODCIIndexFetch(nrows IN NUMBER, rids OUT sys.ODCIRidList, 
    env sys.ODCIEnv) 
  RETURN NUMBER AS
    v_cnum integer;  
    v_rlist sys.ODCIRidList := sys.ODCIRidList();
    idx integer := 1;
    v_rowid varchar2(200);
  BEGIN
    /* TODO implementation required */
    dbms_output.put_line('ODCIIndexFetch>>>>>');
    
    v_cnum := self.curnum;
    loop
      exit when dbms_sql.fetch_rows(v_cnum) = 0;
      v_rlist.extend;
      dbms_sql.column_value(v_cnum, 1, v_rowid);
      dbms_output.put_line('ROWID: '||v_rowid);
      v_rlist(idx) := v_rowid;
      idx := idx +1;
    end loop;
    v_rlist.extend;
    v_rlist(idx) := null;
    rids := v_rlist;
    RETURN ODCIConst.success;
  END ODCIIndexFetch;

  member FUNCTION ODCIIndexClose(env sys.ODCIEnv) RETURN NUMBER AS
    v_cnum integer;
  BEGIN
    /* TODO implementation required */
    v_cnum := self.curnum;
    dbms_sql.close_cursor(v_cnum);
    RETURN ODCIConst.success;
  END ODCIIndexClose;

END;
/

---=================================================================
/* create index type */


create indextype pathindextype for path_between(t_path, float, float) 
using pathindextypemethods;
/

--====================================================================
/* using created indextype in paths table */

create index idx_pathtype on paths(path) indextype is pathindextype;
/


select * from idx_tables;

select * from paths;

select * from paths where path_between(path, 40, 50)=1;
 
insert into paths values (
  'path1', 
  'test path',
  t_path(t_point_list(t_point(0,1), t_point(34, 23), t_point(10, 44)))
); 

commit;  

update  paths set path = t_path(t_point_list(t_point(0,1), t_point(0, 23), t_point(10, 44))) 
  where rowid = 'AAASitAAEAAAAJ0AAD'; 

commit;

--====================================================================
/* creating user defined aggreate function 
    function name: sumpathlen(path t_path) -> 
        will aggregate all t_path lengths
*/

-- define aggregate function implemtation type 
-- object type definition
create or replace
TYPE UAGGFUNCTYPE AS OBJECT 
( /* TODO enter attribute and method declarations here */ 
  agg_val float,
  static function ODCIGetInterfaces(ifclist OUT SYS.ODCIObjectList) RETURN NUMBER,
  STATIC FUNCTION ODCIAggregateInitialize(actx IN OUT UAGGFUNCTYPE) RETURN NUMBER,
  MEMBER FUNCTION ODCIAggregateIterate(self IN OUT UAGGFUNCTYPE, val t_path) RETURN NUMBER,
  MEMBER FUNCTION ODCIAggregateMerge(self IN OUT UAGGFUNCTYPE, ctx2 IN UAGGFUNCTYPE) RETURN NUMBER,
  MEMBER FUNCTION ODCIAggregateTerminate(ReturnValue OUT float, flags IN number) RETURN NUMBER
);
/

-- object type body
create or replace
TYPE BODY UAGGFUNCTYPE AS

  STATIC FUNCTION ODCIAggregateInitialize(actx IN OUT UAGGFUNCTYPE) RETURN NUMBER AS
    v_agg_val float :=0;
  BEGIN
    /* TODO implementation required */
    actx := UAGGFUNCTYPE(v_agg_val);
    RETURN ODCIConst.Success;
  END ODCIAggregateInitialize;

  MEMBER FUNCTION ODCIAggregateIterate(self in out UAGGFUNCTYPE, val t_path) RETURN NUMBER AS
    v_agg_val float;
  BEGIN
    /* TODO implementation required */
    self.agg_val := self.agg_val + val.get_length;
    RETURN ODCIConst.Success;
  END ODCIAggregateIterate;

  MEMBER FUNCTION ODCIAggregateMerge(self in out UAGGFUNCTYPE, ctx2 IN UAGGFUNCTYPE) RETURN NUMBER AS
  BEGIN
    /* TODO implementation required */
    self.agg_val := self.agg_val + ctx2.agg_val;
    RETURN ODCIConst.Success;
  END ODCIAggregateMerge;

  MEMBER FUNCTION ODCIAggregateTerminate(ReturnValue OUT float, flags IN number) RETURN NUMBER AS
  BEGIN
    /* TODO implementation required */
    ReturnValue := self.agg_val;
    RETURN ODCIConst.Success;
  END ODCIAggregateTerminate;

END;
/

-- define aggregate function
CREATE FUNCTION sumpathlen (p t_path) RETURN float
PARALLEL_ENABLE AGGREGATE USING uaggfunctype;
/

-- check aggregate function
select path_name, sumpathlen(path) over (partition by path_name) from paths;